import numpy as np
import scipy.spatial
import gurobipy as gurobi

from .config import CONFIG
from .dtgw_tools import NullObserver, info, shortest_warp_path


def create_gurobi_model():
	model = gurobi.Model()
	if CONFIG["verbosity"] < 1:
		model.setParam("outputFlag", 0)
	model.setParam("MIPGap", CONFIG["gaplimit"] )
	model.setParam("TimeLimit", CONFIG["timelimit"])
	return model


def compute_dtgw(features1, features2, eps, metric=scipy.spatial.distance.cdist, observer=NullObserver()):
	"""Computes the exact dtgw distance between two temporal graphs using
	quadratic programming.
	
	Arguments:
	----------
	features1 : numpy.ndarray of shape (l1, n1, k)
		Features of the first graph.
		features1[i,j,:] is the feature vector
		of vertex j in layer i of the first temporal graph
	features2 : numpy.ndarray of shape (l2, n2, k)
		Analogous to features1
	eps : numpy.ndarray of shape (k)
		Feature vector of a nonexisting (deleted) vertex.
	metric :
		Metric function to compute distances
	observer : 
		Observer object to notify on interal state changes

	Returns:
	---------
	d : float
		The dtgw distance as determined by the integer program solver.
	"""
	
	t1, n1, k1 = features1.shape
	t2, n2, k2 = features2.shape
	if k1 != k2:
		raise ValueError("Arguments have wrong dimensions")
	k = k1
	model = create_gurobi_model()
	features1 = np.hstack((features1, np.broadcast_to(eps, (t1, 1, k))))
	features2 = np.hstack((features2, np.broadcast_to(eps, (t2, 1, k))))

	m = model.addVars(np.ndindex(n1+1, n2+1), vtype=gurobi.GRB.BINARY, name="m")
	p = model.addVars(np.ndindex(t1, t2), vtype=gurobi.GRB.BINARY, name="w")

	if CONFIG["linearize"]:
		z = model.addVars(np.ndindex(t1, t2, n1+1, n2+1), vtype=gurobi.GRB.BINARY, name="z")
		model.addConstrs(
			(z[i,j,k,l] >= w[i,j] + m[k,l] - 1 for i,j,k,l in np.ndindex(t2, t2, n1+1, n2+1)),
			name = "z-constraints"
		)

	model.addConstrs(
		(m.sum(v, "*") == 1 for v in range(n1)),
		name = "mapping-constraints-1"
	)
	model.addConstrs(
		(m.sum("*", w) == 1 for w in range(n2)),
		name = "mapping-constraints-2"
	)

	model.addConstr(p[0, 0] == 1)
	model.addConstrs(
		(p[i, j] <= p[i+1, j+1] + p[i+1 ,j] + p[i, j+1] for i,j in np.ndindex(t1-1, t2-1)),
		name = "warping-constraints-1"
	)
	model.addConstrs(
		(p[i, t2-1] <= p[i+1, t2-1] for i in range(t1-1)),
		name = "warping-constraints-2"
	)
	model.addConstrs(
		(p[t1-1, j] <= p[t1-1, j+1] for j in range(t2-1)),
		name = "warping-constraints-3"
	)

	# initialize warping:
	for var in p.values():
		var.start = 0
	for i, j in shortest_warp_path(t1, t2):
		p[i,j].start = 1

	info("Building objective")
	objective = 0
	for i,j in np.ndindex(t1, t2):
		cost = metric(features1[i,:,:], features2[j,:,:])
		if CONFIG["linearize"]:
			objective += sum(
				z[i,j,v,w] * c
				for (v,w), c in np.ndenumerate(cost)
			)
		else:
			objective += sum(
				p[i,j] * m[v,w] * c
				for (v,w), c in np.ndenumerate(cost)
			)
		if CONFIG["verbosity"] >= 1 and j == 0:
			print("{}/{}".format(i+1, t1))
	model.setObjective(objective, gurobi.GRB.MINIMIZE)
	model.optimize()

	observer.update(
		vertex_matching = (
			(
				v if v < n1 else None,
				w if w < n2 else None
			)
			for (v, w), value in m.items()
			if value.x > 0.01
		),
		warp_path = (
			(i, j)
			for (i, j), value in p.items()
			if value.x > 0.01
		),
		cost = model.objVal,
		final = True
	)

	return model.objVal
